# jarpc - just another RPC
# Copyright (C) 2019  Eugene Ershov
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import warnings

from typing import Any, Dict, Optional

from .abc import ABCServer
from .enums import StatusCode
from .constants import NoValue


class Request:

    __slots__ = (
        "server",
        "command_index",
        "node",
        "_data",
        "_address",
        "_reply_called",
    )

    def __init__(
        self,
        server: ABCServer,
        command_index: int,
        node: str,
        data: Any,
        address: Optional[str],
    ):
        self.server = server

        self.command_index = command_index
        self.node = node

        self._data = data
        self._address = address

        self._reply_called = False

    @classmethod
    def from_data(cls, server: ABCServer, payload: Dict[str, Any]) -> "Request":
        return cls(
            server=server,
            command_index=payload["c"],
            node=payload["n"],
            data=payload.get("d", {}),
            address=payload.get("a"),
        )

    async def reply(self, data: Any) -> None:
        await self._reply_with_status(data)

    async def _reply_with_status(
        self, data: Any = NoValue, status: StatusCode = StatusCode.SUCCESS
    ) -> None:
        if self._reply_called:
            warnings.warn(
                "Reply function was called already. Using it multiple times may cause problems"
            )
        else:
            self._reply_called = True
            await self.server.reply(address=self._address, data=data, status=status)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} command_index={self.command_index}>"



# jarpc - just another RPC
# Copyright (C) 2019  Eugene Ershov
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Dict

from .enums import StatusCode


class Response:

    __slots__ = ("status", "node", "data", "_address")

    def __init__(self, status: StatusCode, node: str, data: Any, address: str):
        self.status = status
        self.node = node
        self.data = data

        self._address = address

    @classmethod
    def from_data(cls, payload: Dict[str, Any]) -> "Response":
        return cls(
            status=StatusCode(payload["s"]),
            node=payload["n"],
            data=payload.get("d", {}),
            address=payload["a"],
        )

    def __str__(self) -> str:
        return str(self.data)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} status={self.status} node={self.node} data={self.data}>"



# jarpc - just another RPC
# Copyright (C) 2019  Eugene Ershov
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

from typing import Any, Dict, Callable, Optional

from .abc import ABCServer
from .enums import StatusCode
from .typing import CommandType
from .request import Request
from .constants import NoValue
from .connection import Connection

log = logging.getLogger(__name__)


class Server(Connection, ABCServer):
    """Listens for commands from clients and sends responses."""

    # NOTE: defining different __slots__ in Client and Server causes error creating
    # Slient

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self._commands: Dict[int, CommandType] = {}

    def command(self, index: int) -> Callable[[CommandType], None]:
        """Flask-style decorator used to register commands. Calls register_command."""

        def inner(func: CommandType) -> None:
            self.add_command(index, func)

        return inner

    def add_command(self, index: int, fn: CommandType) -> int:
        """Registers new command. Raises ValueError if index already used."""

        if index in self._commands:
            raise ValueError("Command with index %d already registered", index)

        self._commands[index] = fn

        return index

    def remove_command(self, index: int) -> CommandType:
        """Removes existing command. Raises ValueError if index not found."""

        if index not in self._commands:
            raise ValueError("Command with index %d is not registered", index)

        return self._commands.pop(index)

    def _make_request(self, data: Any) -> Optional[Request]:
        return Request.from_data(self, data)

    async def _handle_request(self, request: Request) -> None:
        log.info("received command %d", request.command_index)

        fn = self._commands.get(request.command_index)
        if fn is None:
            log.warning("unknown command %d", request.command_index)

            await request._reply_with_status(status=StatusCode.UNKNOWN_COMMAND)

            return

        try:
            command = fn(request, **request._data)
        except TypeError as e:
            log.error("bad arguments given to %d: %s", request.command_index, str(e))

            await request._reply_with_status(str(e), StatusCode.BAD_PARAMS)

            return

        try:
            command_result = await command
        except Exception as e:
            log.error(
                "error calling command %d %s: %s",
                request.command_index,
                e.__class__.__name__,
                str(e),
            )

            await request._reply_with_status(str(e), StatusCode.INTERNAL_ERROR)

            return

        if command_result is None:
            # Special case, should be documented.
            # Returning None is allowed using request.reply
            return

        await request.reply(command_result)

    async def reply(
        self, *, address: Optional[str], status: StatusCode, data: Any
    ) -> None:
        """Sends response to address (if address is present)."""

        if address is None:
            log.debug("no address, unable to respond")
            return

        payload = {"s": status.value, "n": self.node, "a": address}

        if data is not NoValue:
            payload["d"] = data

        await self._send_response(payload)



# jarpc - just another RPC
# Copyright (C) 2019  Eugene Ershov
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import time
import uuid
import asyncio
import logging

from typing import Any, Dict, List, Optional, Generator

from .abc import ABCClient, ResponsesIterator
from .typing import TypedQueue
from .response import Response
from .connection import Connection

log = logging.getLogger(__name__)


class ResponsesWithTimeout(ResponsesIterator):
    """Provides access to command responses for limited time."""

    __slots__ = (
        "_client",
        "_address",
        "_timeout",
        "_expect_responses",
        "_responses_seen",
        "_start_time",
        "_queue",
    )

    def __init__(
        self,
        client: "Client",
        queue: TypedQueue[Response],
        address: str,
        timeout: float,
        expect_responses: Optional[int] = None,
    ):
        self._client = client
        self._queue = queue
        self._address = address
        self._timeout = timeout

        if expect_responses is None:
            expect_responses = 0

        if expect_responses < 0:
            raise ValueError("expect_responses should be >= 0")

        self._expect_responses = expect_responses

        self._responses_seen = 0

        self._start_time = time.time()

    def __await__(self) -> Generator[Any, None, List[Response]]:
        """Returns all responses once they are ready."""

        async def coro() -> List[Response]:
            return [resp async for resp in self]

        return coro().__await__()

    def __aiter__(self) -> "ResponsesWithTimeout":
        return self

    async def __anext__(self) -> Response:
        if (
            self._expect_responses != 0
            and self._expect_responses == self._responses_seen
        ):
            raise StopAsyncIteration

        try:
            resp = await asyncio.wait_for(
                self._queue.get(), timeout=self.time_remaining, loop=self._client._loop
            )
        except asyncio.TimeoutError:
            raise StopAsyncIteration

        self._responses_seen += 1

        return resp

    def __del__(self) -> None:
        self._client._remove_queue(self._address)

    @property
    def responses_seen(self) -> int:
        """Amount of yielded responses."""

        return self._responses_seen

    @property
    def time_remaining(self) -> float:
        """Remaining time until iterator is closed."""

        return self._timeout - time.time() + self._start_time

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} time_remaining={self.time_remaining} responses_seen={self.responses_seen}>"


class EmptyResponses(ResponsesIterator):
    """Behaves the same way as ResponsesWithTimeout, except it is empty."""

    def __await__(self) -> Generator[Any, None, List[Response]]:
        """Returns empty list of responses."""

        async def coro() -> List[Response]:
            return []

        return coro().__await__()

    def __aiter__(self) -> "EmptyResponses":
        return self

    async def __anext__(self) -> Response:
        raise StopAsyncIteration


class Client(Connection, ABCClient):
    """Sends commands to servers and listens for responses."""

    # NOTE: defining different __slots__ in Client and Server causes error creating
    # Slient

    def __init__(
        self,
        *args: Any,
        default_timeout: Optional[int] = None,
        default_expect_responses: Optional[int] = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)

        self._default_timeout = default_timeout
        self._default_expect_responses = default_expect_responses

        self._listeners: Dict[str, TypedQueue[Response]] = {}

    def _make_response(self, data: Any) -> Optional[Response]:
        return Response.from_data(data)

    async def _handle_response(self, response: Response) -> None:
        log.info("received response from node %s", response.node)

        queue = self._listeners.get(response._address)
        if queue is None:
            log.debug("ignoring response from node %s", response.node)
            return

        await queue.put(response)

    def _add_queue(self, address: str, queue: TypedQueue[Response]) -> None:
        self._listeners[address] = queue

    def _remove_queue(self, address: str) -> None:
        self._listeners.pop(address, None)

    def call(
        self,
        command_index: int,
        data: Dict[str, Any] = {},
        timeout: Optional[float] = None,
        expect_responses: Optional[int] = None,
    ) -> ResponsesIterator:
        """
        Calls command and returns received responses. Skips response processing
        completely if timeout is None.
        """

        log.info("sending command %d", command_index)

        payload = {"n": self._node, "c": command_index, "d": data}

        if timeout is None:
            timeout = self._default_timeout

        if timeout is not None:
            address = uuid.uuid4().hex
            payload["a"] = address

            queue: TypedQueue[Response] = TypedQueue(loop=self._loop)
            self._add_queue(address, queue)

        self._loop.create_task(self._send_request(payload))

        if timeout is None:
            return EmptyResponses()
        else:
            return ResponsesWithTimeout(
                self,
                queue,
                address,
                timeout,
                expect_responses or self._default_expect_responses,
            )



# jarpc - just another RPC
# Copyright (C) 2019  Eugene Ershov
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio

from typing import Any, Union, Generic, TypeVar, Callable

__all__ = ("Serializer", "Deserializer", "CommandType")

Serializer = Union[Callable[[Any], bytes], Callable[[Any], str]]
Deserializer = Callable[[bytes], Any]

CommandType = Any

# TODO: annotate _CommandType properly.
# Possible solutions (do not work for different reasons)
#
# typing_extensions.Protocol
# class _CommandType(Protocol):
#     def __call__(self, __request: Request, **kwargs: Any) -> Any:
#         ...
#
# mypy_extensions.KwArg
# _CommandType = Callable[["Server", KwArg(Any)], Awaitable[Any]]
#
# typing.TypeVar ?


T = TypeVar("T")


class TypedQueue(asyncio.Queue, Generic[T]):  # type: ignore
    ...



from jarpc import StatusCode

all_enumaration_types = [
    StatusCode.SUCCESS,
    StatusCode.BAD_FORMAT,
    StatusCode.UNKNOWN_COMMAND,
    StatusCode.BAD_PARAMS,
    StatusCode.INTERNAL_ERROR,
]


def test_properties_values():
    """
    Should ensure that variables are set to their expected values
    """

    vals = [0, 1, 2, 3, 4]
    for enum in all_enumaration_types:
        assert enum.value == vals[enum.value]


def test_properties_types():
    """
    Should have the enum variable of type Enum
    """

    for enum in all_enumaration_types:
        assert type(enum) == StatusCode



# jarpc - just another RPC
# Copyright (C) 2019  Eugene Ershov
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import abc

from typing import Any, Dict, List, Tuple, Union, Optional, Generator

from .enums import StatusCode
from .typing import CommandType
from .response import Response

__all__ = ("ResponsesIterator", "ABCConnection", "ABCClient", "ABCServer")


class ResponsesIterator(abc.ABC):
    """
    Provides access to command responses.

    Instances can be awaited to get all responses at once or used as async iterator to
    process responses as soon as they come.
    """

    @abc.abstractmethod
    def __await__(self) -> Generator[Any, None, List[Response]]:
        """Returns all responses once they are ready."""

    @abc.abstractmethod
    def __aiter__(self) -> "ResponsesIterator":
        ...

    @abc.abstractmethod
    async def __anext__(self) -> Response:
        ...


class ABCConnection(abc.ABC):
    """RPC Connection."""

    @abc.abstractproperty
    def name(self) -> str:
        """Connection name."""

    @abc.abstractproperty
    def node(self) -> str:
        """Node identifier."""

    @abc.abstractmethod
    async def start(
        self, redis_address: Union[Tuple[str, int], str], **kwargs: Any
    ) -> None:
        """Starts processing messages."""

    @abc.abstractmethod
    def close(self) -> None:
        """Closes connection."""


class ABCClient(ABCConnection):
    """Calls commands."""

    @abc.abstractmethod
    def call(
        self,
        command_index: int,
        data: Dict[str, Any] = {},
        timeout: Optional[float] = None,
        expect_responses: int = 0,
    ) -> ResponsesIterator:
        """Calls command by index."""


class ABCServer(ABCConnection):
    """Responds to commands."""

    @abc.abstractmethod
    def add_command(self, index: int, fn: CommandType) -> int:
        """Registers new command."""

    @abc.abstractmethod
    def remove_command(self, index: int) -> CommandType:
        """Removes existing command."""

    @abc.abstractmethod
    async def reply(
        self, *, address: Optional[str], status: StatusCode, data: Any
    ) -> None:
        """Sends response to address."""



from .client import Client
from .server import Server
from .request import Request


class Slient(Client, Server):
    """Slient combines functionality of Client and Server."""

    async def _handle_request(self, request: Request) -> None:
        if request.node == self.node:
            # ignore requests sent from this slient
            return

        await super()._handle_request(request)



import re

import setuptools  # type: ignore

install_requires = ["aioredis"]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

version = ""
author = ""

with open("jarpc/__init__.py") as f:
    cont = f.read()
    version_gr = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', cont, re.MULTILINE
    )
    if version_gr:
        version = version_gr.group(1)
    author_gr = re.search(r'^__author__\s*=\s*[\'"]([^\'"]*)[\'"]', cont, re.MULTILINE)
    if author_gr:
        author = author_gr.group(1)

setuptools.setup(
    name="jarpc",
    version=version,
    author=author,
    author_email="fogaprod@gmail.com",
    description="RPC over redis communication library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IOMirea/jarpc",
    install_requires=install_requires,
    python_requires=">=3.6",
    packages=setuptools.find_packages(),
    license="GPLv3",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Topic :: Software Development :: Object Brokering",
        "Typing :: Typed",
    ],
)



# jarpc - just another RPC
# Copyright (C) 2019  Eugene Ershov
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import uuid
import asyncio
import logging

from typing import Any, Tuple, Union, Mapping, Optional

import aioredis

from .abc import ABCConnection
from .enums import MessageType
from .typing import Serializer, Deserializer
from .request import Request
from .response import Response

log = logging.getLogger(__name__)


_payload_value_to_type = {b"0": MessageType.REQUEST, b"1": MessageType.RESPONSE}
_payload_type_to_value = {v: k for k, v in _payload_value_to_type.items()}


class Connection(ABCConnection):

    __slots__ = ("_name", "_node" "_loop", "_ready", "_loads", "_dumps", "_sub", "_pub")

    def __init__(
        self,
        name: str,
        loads: Optional[Deserializer] = None,
        dumps: Optional[Serializer] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        node: Optional[str] = None,
    ):
        self._name = f"rpc:{name}"
        self._node = uuid.uuid4().hex if node is None else node

        self._loop = asyncio.get_event_loop() if loop is None else loop
        self._ready = asyncio.Event()

        if loads and dumps:
            self._loads = loads
            self._dumps = dumps
        elif loads is None and dumps is None:
            import marshal

            self._loads = marshal.loads
            self._dumps = marshal.dumps
        else:
            raise ValueError("You cannot define only one of dumps and loads.")

    async def start(
        self, redis_address: Union[Tuple[str, int], str], **kwargs: Any
    ) -> None:
        """Starts processing messages."""

        self._sub = await aioredis.create_redis(
            redis_address, loop=self._loop, **kwargs
        )
        self._pub = await aioredis.create_redis(
            redis_address, loop=self._loop, **kwargs
        )

        channels = await self._sub.subscribe(self._name)
        assert len(channels) == 1

        self._ready.set()

        log.info(f"sub: connected: {self._name}")
        log.info(f"pub: connected: {self._name}")

        await self._handler(channels[0])

    def run(self, *args: Any, **kwargs: Any) -> None:
        """
        A blocking way to launch message processing.
        Takes same arguments as Connection.start.
        """

        self._loop.run_until_complete(self.start(*args, **kwargs))

    async def wait_until_ready(self) -> None:
        """Waits until connection is established."""

        await self._ready.wait()

    async def _handler(self, channel: aioredis.pubsub.Channel) -> None:
        async for msg in channel.iter():
            # TODO: parser
            message_type_byte = msg[0:1]
            msg_type = _payload_value_to_type.get(
                message_type_byte, MessageType.UNKNOWN_TYPE
            )
            if msg_type == MessageType.UNKNOWN_TYPE:
                log.warn(f"Unknown payload type: {message_type_byte}")
                continue

            try:
                data = self._loads(msg[1:])
            except Exception:
                # address is unavailable
                # await self._respond(request.address, StatusCode.BAD_FORMAT)

                log.exception(
                    "could not deserialize payload. Make sure to use same algorithm on both ends"
                )
                continue

            if msg_type == MessageType.REQUEST:
                request = self._make_request(data)
                if request is not None:
                    await self._handle_request(request)

            elif msg_type == MessageType.RESPONSE:
                response = self._make_response(data)
                if response is not None:
                    await self._handle_response(response)

    def _make_request(self, data: Any) -> Optional[Request]:
        """Called for creating request from data. Overridable."""

        return None

    def _make_response(self, data: Any) -> Optional[Response]:
        """Called for creating response from data. Overridable."""

        return None

    async def _handle_request(self, request: Request) -> None:
        """Called for handling request. Overridable."""

        pass

    async def _handle_response(self, response: Response) -> None:
        """Called for handling response. Overridable."""

        pass

    async def _send_request(self, payload: Mapping[str, Any]) -> None:
        """Sends payload of request type."""

        await self._send(payload, _payload_type_to_value[MessageType.REQUEST])

    async def _send_response(self, payload: Mapping[str, Any]) -> None:
        """Sends payload of response type."""

        await self._send(payload, _payload_type_to_value[MessageType.RESPONSE])

    async def _send(self, payload: Mapping[str, Any], pl_type: bytes) -> None:
        """Sends payload of given type."""

        encoded = self._dumps(payload)
        if isinstance(encoded, str):
            encoded = encoded.encode()

        encoded = pl_type + encoded

        num_listeners = await self._pub.publish(self._name, encoded)
        log.debug(f"delivered to {num_listeners} listeners")

    def close(self) -> None:
        """Closes connection."""

        log.info("closing connections")

        self._ready.clear()

        self._sub.close()
        self._pub.close()

    @property
    def name(self) -> str:
        """Connection name."""

        return self._name

    @property
    def node(self) -> str:
        """Node identifier."""

        return self._node

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name} node={self.node}>"
